#include "parser.h"
#include "scanner.h"
#include "nativedescparser.h"
#include "nativedescscanner.h"
#include "options.h"
#include "compilerexc.h"

#include <cstdio>
#include <fstream>

using namespace std;

static const char * const optv[] = {
	"t:target",
	"l?list",
	"o:output",
	"I:",
	NULL,
};

int main(int argc,char **argv)
{
	int optchar;
	const char *optarg;
	vector<string> includePath;

	bool makeListing = false;
	string possibleFileName = "";
	string targetName;
	string outputName;
	string listFileName("");
	string sourceFileName;
	int optErrors = 0;


	Options  opts(*argv, optv);
    opts.ctrls(Options::PARSE_POS);
	OptIter  *iter = new OptArgvIter((--argc, ++argv));
	
	while( (optchar = opts(*iter, optarg)) ) {
		switch (optchar) {
			case 'I':
				if (optarg) includePath.push_back(optarg);
				else optErrors++;
				break;
			case 't':
				if (optarg) targetName = optarg;
				else optErrors++;
				possibleFileName = "";
				break;
			case 'o':
				if (optarg) outputName = optarg;
				else optErrors++;
				possibleFileName = "";
				break;
			case 'l':
				if (optarg) {
					listFileName = optarg;
					possibleFileName = optarg;
				}
				makeListing = true;
				break;
			case Options::POSITIONAL:
				sourceFileName = optarg;
				possibleFileName = "";
				break;
			case Options::BADCHAR:
				optErrors++;
				break;
		}
	}

	if (optErrors) {
		opts.usage(cerr, "file");
		exit(1);
	}

	if (sourceFileName == "") {
		listFileName = "";
	}

	if (targetName == "")  {
		cerr << "No target specified\n";
		return 1;
	}

	ifstream tgt;
	tgt.open(targetName.c_str());
	if (!tgt.is_open()) {
		cerr << "Cannot open target description file\n";
		return 1;
	}
	NativeDescParser tgtParser(&tgt);
	int result = tgtParser.parse();
	if (result) {
		cerr << "Error parsing target description\n";
		return 1;
	}

	TargetDesc targetDesc = tgtParser.result();
	Program::instance()->context()->target = &targetDesc;
	if (targetDesc.ram <= 0 || targetDesc.ram > 32768 ) {
		cerr << "Invalid target RAM size\n";
		return 1;
	}
	
	if (targetDesc.program <= 0 || targetDesc.program > 32768 ) {
		cerr << "Invalid target program memory size\n";
		return 1;
	}

	if (possibleFileName != "" && sourceFileName == "") {
		sourceFileName = possibleFileName;
	}
	if (sourceFileName == "") {
		opts.usage(cerr, "file");
		exit(1);
	}

	ifstream sourceStream(sourceFileName.c_str());
	if (!sourceStream.is_open()) {
		cerr << "Failed to open source file " << sourceFileName <<'\n';
		return 1;
	}

	if (listFileName == "") {
		listFileName = sourceFileName + ".lst";
	}
	
	if (outputName == "") {
		outputName = sourceFileName + ".bin";
	}

    Parser aCompiler(&sourceStream, &sourceFileName);
	aCompiler.setIncludePath(includePath);
	try {
	    int result = aCompiler.parse();
	


		if (result) {
			return 1;
		}

		Program::instance()->fold();
		//cout << *(Program::instance());
	
		CodeBlock *code = Program::instance()->compile();

		if (makeListing) {
			ofstream ls;
			ls.open(listFileName.c_str());
			if (!ls.is_open()) {
				cerr << "Failed to open list file " << listFileName <<'\n';
				return 1;
			}
			ls << *code;
		}

		ofstream os;
		os.open(outputName.c_str());
		if (!os.is_open()) {
			cerr << "Failed to open output file " << outputName <<'\n';
			return 1;
		}
		code->generate(os);

	} catch (string e) {
		cerr << e << "\n";
		return 1;
	} catch (CompilerException e) {
		cerr << e.message() << " in " << e.filename() <<  " line " << e.line() <<  '\n';
		return 1;
	}

    return 0;

	
};

